/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package pogr1_holec;

import java.util.*;
//import java.awt.*;
/**
 *The implementation of 3D convex hull algorithm
 * @author milan
 */
public class ConvexHull {
//==============================================================================
// Define inner classes
// Classes to be used within Convexhull only
// Original class for ConvexHullElement of set of points in n-dimensional space
    private class HullElement {
        // Data
        public int[] visiblePoints;
        public int furthestPointIndex;
        public point center;
        protected double[] hyperplane;
        protected point[] node;
        protected point[][] edge;
        protected HullElement[] neighbour;
        // Methods
        protected HullElement FindVisibleNeighbours(int _furthestPoint, point _center) {
            return this;
        }

        protected void changeNeighbour(HullElement elementNew, point p1, point p2) {

        }

        public HullElement() {
            visiblePoints = new int[2*NumberOfPoints];
        }
    }
//==============================================================================
// Points from 2D space
    private class HullElement2D extends HullElement {
        // Data
        
        // Methods
        private void setHyperplane(point A, point B, point Center) {
            hyperplane[0] = A.getY() - B.getY();
            hyperplane[1] = B.getX() - A.getX();
            hyperplane[2] = hyperplane[0]*A.getX() + hyperplane[1]*A.getY();
            if( positionToHyperplane(Center) > hyperplane[2] ) {
                hyperplane[0] *= -1;
                hyperplane[1] *= -1;
                hyperplane[2] *= -1;
            }
        }

        private double positionToHyperplane(point point) {
            return hyperplane[0]*point.getX() + hyperplane[1]*point.getY();
        }

        private int FindVisiblePoints() {
            int furthest = 0;
            double distance = hyperplane[2];
            for(int i=1; i<=NumberOfPoints; ++i) {
                if(positionToHyperplane(SetOfPoints[i]) < hyperplane[2]) {
                    visiblePoints[i] = 1;
                    continue;
                }
                if(positionToHyperplane(SetOfPoints[i]) > distance) {
                    distance = positionToHyperplane(SetOfPoints[i]);
                    furthest = i;
                }
            }
            return furthest;
        }

        @Override protected HullElement FindVisibleNeighbours(int _furthestPoint, point _center) {
            for(int i = 0; i < 2; ++i) {
                if( neighbour[i].visiblePoints[_furthestPoint] == 0)
                    FindVisibleNeighbours(_furthestPoint, _center);
                else {
                    HullElement a = new HullElement2D(edge[i][0], SetOfPoints[_furthestPoint], _center);
                    addElement(a);
                    neighbour[i].changeNeighbour(a, edge[i][0], edge[i][1]);
                    a.changeNeighbour(neighbour[i], edge[i][0], edge[i][1]);
                }
            }
            return this;
        }

        @Override protected void changeNeighbour(HullElement elementNew, point p1, point p2) {
            for(int i=0; i < 2; ++i)
                if(p1==edge[i][0]) neighbour[i] = elementNew;
        }

        public HullElement2D(point A, point B, point Center) {
            node = new point[2];
            hyperplane = new double[3];
            edge = new point[1][2];
            neighbour = new HullElement2D[2];
            center = new point((A.getX()+B.getX())/2, (A.getY()+B.getY())/2, 0);
            init(A, B, Center);
        }

        private void init(point A, point B, point Center) {
            edge[0][0] = A;
            edge[1][0] = B;
            setHyperplane(A, B, Center);
            furthestPointIndex = 1;
        }
    }
//==============================================================================
// Points from 3D space
// HullElement means a linear piece of a hyperplane dividing the all space into two parts(space => points)
    private class HullElement3D extends HullElement {
        // Data
       
        // Methods
        /** setHyperplane(), function computing geometric expression of hullelement */
        private void setHyperplane(point A, point B, point C, point Center) {
            double[] a = new double[3];
            double[] b = new double[3];
            a[0] = B.getX() - A.getX();
            a[1] = B.getY() - A.getY();
            a[2] = B.getZ() - A.getZ();

            b[0] = C.getX() - A.getX();
            b[1] = C.getY() - A.getY();
            b[2] = C.getZ() - A.getZ();

            hyperplane[0] = a[1]*b[2] - a[2]*b[1];
            hyperplane[1] = a[2]*b[0] - a[0]*b[2];
            hyperplane[2] = a[0]*b[1] - a[1]*b[0];
            hyperplane[3] = hyperplane[0]*A.getX() + hyperplane[1]*A.getY() + hyperplane[2]*A.getZ();
            if( positionToHyperplane(Center) > hyperplane[3] ) {
                hyperplane[0] *= -1;
                hyperplane[1] *= -1;
                hyperplane[2] *= -1;
                hyperplane[3] *= -1;
            }
        }
        /** positionToHyperplane(), function determinig position of point to the hullelement(upvisible or down visible) */
        private double positionToHyperplane(point point) {
            return hyperplane[0]*point.getX() + hyperplane[1]*point.getY() + hyperplane[2]*point.getZ();
        }

        protected double[] OrientationOfHyperplane() {
            double[] direction = new double[4];
            for(int i=0; i<4; ++i)
                direction[i] = hyperplane[i];
            return direction;
        }

        /** FindVisiblePoints(), function computing "upvisible points"(means points outside the current convex hull),
         * return is the furthest point from hullelement */
        private int FindVisiblePoints() {
            int furthest = 0;
            double distance = hyperplane[3];
            for(int i=0; i<RemainingPoints.size(); ++i) {
                if(positionToHyperplane(SetOfPoints[RemainingPoints.get(i)]) <= hyperplane[3]) {
                    visiblePoints[RemainingPoints.get(i)] = 1;
                    continue;
                }
                if(positionToHyperplane(SetOfPoints[RemainingPoints.get(i)]) > distance) {
                    distance = positionToHyperplane(SetOfPoints[RemainingPoints.get(i)]);
                    furthest = RemainingPoints.get(i);
                }
            }
            return furthest;
        }

        /** cahngeNeighbour(), function changing neighbour of a hullelement connected with another hullelement through edge */
        @Override protected void changeNeighbour(HullElement elementNew, point p1, point p2) {
            for(int i=0; i < 3; ++i)
                if( ((p1==edge[i][0]) && (p2==edge[i][1])) || ((p1==edge[i][1]) && (p2==edge[i][0])) )
                    neighbour[i] = elementNew;
        }

        @Override protected HullElement FindVisibleNeighbours(int _furthestPointIndex, point _center) {
            for(int i = 0; i < 3; ++i) {
                if(neighbour[i] == null)
                    continue;
                if( neighbour[i].visiblePoints[_furthestPointIndex] == 0 ) {
                    HullElement temp = neighbour[i];
                    neighbour[i] = null;
                    temp.FindVisibleNeighbours(_furthestPointIndex, _center);
                    ElementsToExecute.remove(temp);
                }
                else {
                    createNewElement(neighbour[i], edge[i], _furthestPointIndex, _center);
                    neighbour[i] = null;
                }
            }
            return this;
        }

        public HullElement3D(point A, point B, point C, point Center) {
            node = new point[3];
            hyperplane = new double[4];
            edge = new point[3][2];
            neighbour = new HullElement3D[3];
            init(A, B, C, Center);
        }

        private void init(point A, point B, point C, point Center) {
            center = new point((A.getX()+B.getX()+C.getX())/3,(A.getY()+B.getY()+C.getY())/3,(A.getZ()+B.getZ()+C.getZ())/3);
            node[0] = A;
            node[1] = B;
            node[2] = C;
            setHyperplane(A, B, C, Center);
            furthestPointIndex = FindVisiblePoints();
            if(furthestPointIndex == 0) {
                ConvexHullStack.add(this);
                for(int i=0; i<3; ++i)
                    addConvexHullNode(node[i]);
                addConvexHullEdge(A, B);
                addConvexHullEdge(A, C);
                addConvexHullEdge(C, B);
                addConvexHullPlane(A, B, C);
            }
            edge[0][0] = A;
            edge[0][1] = B;
            edge[1][0] = B;
            edge[1][1] = C;
            edge[2][0] = C;
            edge[2][1] = A;
        }
    }
//==============================================================================
// Data
    private point[] SetOfPoints;
    private int NumberOfPoints;
    private HullElement[][] NewElements;
    private point[] NeighbourPoints;
    private int NumOfNeighbourPoints;
    private ArrayList<Integer> RemainingPoints = new ArrayList<Integer>();
    private ArrayList<HullElement> ConstructedElements = new ArrayList<HullElement>();
    private ArrayList<HullElement> ElementsToExecute = new ArrayList<HullElement>();
    private ArrayList<HullElement> ConvexHullStack = new ArrayList<HullElement>();
    public ArrayList<line> ConvexHullSceleton = new ArrayList<line>();
    public ArrayList<triangle> ConvexHullSurface = new ArrayList<triangle>();
    public ArrayList<point> ConvexHullNodes = new ArrayList<point>();
    public ArrayList<line> ConvexHull2D = new ArrayList<line>();
// Methods
    public ConvexHull(ArrayList<point> InputList) {
        //int n = 100;
        int n = InputList.size();
        if(n<4)
            System.exit(n);
        NewElements = new HullElement [n][10];
        NeighbourPoints = new point[2*n];
        NumOfNeighbourPoints = 0;
        NumberOfPoints = n;
        SetOfPoints = new point[2*n];
        for(int i = 0; i<InputList.size(); ++i)
            SetOfPoints[i+1] = InputList.get(i);
        Init();
    }

    public ConvexHull(ArrayList<point> InputList, int disting) {
        //int n = 100;
        int n = InputList.size();
        if(n<4)
            System.exit(n);
        NewElements = new HullElement [n][10];
        NeighbourPoints = new point[2*n];
        NumOfNeighbourPoints = 0;
        NumberOfPoints = n;
        SetOfPoints = new point[2*n];
        for(int i = 0; i<InputList.size(); ++i)
            SetOfPoints[i+1] = InputList.get(i);
        Init(NumberOfPoints);
    }

    private void Init()  {
        // Generating set of points

/*        int n = 1;
        for(int i=0; i<5; i++) {
            for(int j=0; j<5; j++) {
                for(int k=0; k<5; k++) {
                    SetOfPoints[n] = new point(i,j,k);
                    n++;
                }
            }
        }
        // Initializing original hull
        point A = SetOfPoints[34];
        point B = SetOfPoints[32];
        point C = SetOfPoints[62];
        point T = SetOfPoints[1];
        NumberOfPoints = n-1;
 /*
        point A = new point(1, 0, 0);
        point B = new point(0, 1, 0);
        point C = new point(0, 0, 1);
        point T = new point(0, 0, 0);
        point P1 = new point(4, 4, 4);
        point P2 = new point(-10,-10,-10);
        point P3 = new point(10,0,0);
        point P4 = new point(0,10,0);
        point P5 = new point(0,0,10);
        point P6 = new point(10,10,10);
        point P7 = new point(-1, -2, -3);
        point P8 = new point(-4, -4, -4);
        point P9 = new point(2, 3, 5);
        point P10 = new point(12, 0, 0);
        point P11 = new point(9, 10, -1);
        point P12 = new point(100, 0, 0);
        point P13 = new point(0, 100, 0);
        point P14 = new point(0, 0, 100);
        point P15 = new point(-100, -100, -100);

        SetOfPoints[1]  = A;
        SetOfPoints[2]  = B;
        SetOfPoints[3]  = C;
        SetOfPoints[4]  = T;
        SetOfPoints[5]  = P1;
        SetOfPoints[6]  = P2;
        SetOfPoints[7]  = P3;
        SetOfPoints[8]  = P4;
        SetOfPoints[9]  = P5;
        SetOfPoints[10] = P6;
        SetOfPoints[11] = P7;
        SetOfPoints[12] = P8;
        SetOfPoints[13] = P9;
        SetOfPoints[14] = P10;
        SetOfPoints[15] = P11;
        SetOfPoints[16] = P12;
        SetOfPoints[17] = P13;
        SetOfPoints[18] = P14;
        SetOfPoints[19] = P15;
        NumberOfPoints = 19;
*/

        // Initial points valid for 2DConvexhull
        point A = SetOfPoints[1];
        point B = SetOfPoints[3];
        point C = SetOfPoints[5];
        point Temp = new point(SetOfPoints[1].getX(), SetOfPoints[1].getY(), 2);

        for(int i=1; i<=NumberOfPoints; ++i)
            RemainingPoints.add(i);

        HullElement temp = new HullElement3D(A, B, C, Temp);
        int peakPointIndex = temp.furthestPointIndex;
  
        HullElement E1 = new HullElement3D(A, B, C, SetOfPoints[peakPointIndex]);
        if(E1.furthestPointIndex != 0)
                addElement(E1);
        for(int i = 0; i<3; ++i) {
            createNewElement(E1, E1.edge[i], peakPointIndex, E1.center);
        }
        setNeighbours(peakPointIndex);
        ReducePoints(temp);
    }

    private void Init(int disting)  {
        // Generating initial set of points valid for Delaunay triangulation
        int peakPointIndex = NumberOfPoints;

        point A = SetOfPoints[1];
        point B = SetOfPoints[2];
        point C = SetOfPoints[3];
        point Temp = SetOfPoints[peakPointIndex];

        for(int i=1; i<=NumberOfPoints; ++i)
            RemainingPoints.add(i);

        HullElement E1 = new HullElement3D(A, B, C, Temp);
        if(E1.furthestPointIndex != 0)
                addElement(E1);
        for(int i = 0; i<3; ++i) {
            createNewElement(E1, E1.edge[i], peakPointIndex, E1.center);
        }
        setNeighbours(peakPointIndex);
        //ReducePoints(temp);
    }

    private void QuickHullStep(HullElement element) {
        element.FindVisibleNeighbours(element.furthestPointIndex, element.center);
        setNeighbours(element.furthestPointIndex);
        ReducePoints(element);
        ElementsToExecute.remove(element);
    }

    private void createNewElement(HullElement _neighbour, point[] _edge, int _peakPointIndex, point T) {
        HullElement a = new HullElement3D(_edge[0], _edge[1], SetOfPoints[_peakPointIndex], T);
        _neighbour.changeNeighbour(a, _edge[0], _edge[1]);
        a.changeNeighbour(_neighbour, _edge[0], _edge[1]);
        addNeighbour(a, _edge[0], _edge[1]);
        ConstructedElements.add(a);
        if(a.furthestPointIndex != 0)
                addElement(a);
    }

    private void setNeighbours(int peak) {
        for(int i=0; i<NumOfNeighbourPoints; ++i) {
            NewElements[i][0].changeNeighbour(NewElements[i][1], NeighbourPoints[i], SetOfPoints[peak]);
            NewElements[i][1].changeNeighbour(NewElements[i][0], NeighbourPoints[i], SetOfPoints[peak]);
            NewElements[i][0] = null;
            NewElements[i][1] = null;
            NeighbourPoints[i] = null;
        }
        NumOfNeighbourPoints = 0;
    }

    private void addNeighbour(HullElement elem, point node1, point node2) {
        int n1 = 1;
        int n2 = 1;
        int i = 0;

        for( ; i<NumOfNeighbourPoints; ++i) {
            if(NeighbourPoints[i] == node1) {
                NewElements[i][1] = elem;
                n1 = 0;
            }
            else {
                if(NeighbourPoints[i] == node2) {
                    NewElements[i][1] = elem;
                    n2 = 0;
                }
            }
        }
        if(n1 != 0) {
            NumOfNeighbourPoints = ++i;
            NeighbourPoints[i-1] = node1;
            NewElements[i-1][0] = elem;
        }
        if(n2 != 0) {
            NumOfNeighbourPoints = ++i;
            NeighbourPoints[i-1] = node2;
            NewElements[i-1][0] = elem;
        }
    };

    private void addElement(HullElement element) {
        ElementsToExecute.add(element);
    }

    private void ReducePoints(HullElement rootElement) {
        ArrayList<Integer> pointsToBeRemoved = new ArrayList<Integer>();
        for(Iterator i = RemainingPoints.iterator(); i.hasNext(); ) {
            Integer indexOfPointToCheck = (Integer)i.next();
            if(rootElement.visiblePoints[indexOfPointToCheck] == 0) {
                int product = 1;
                for(int j=0; j<ConstructedElements.size(); ++j) {
                    product *= ConstructedElements.get(j).visiblePoints[indexOfPointToCheck];
                }
                if(product != 0)
                    pointsToBeRemoved.add(indexOfPointToCheck);
            }
        }

        for(Iterator i = pointsToBeRemoved.iterator(); i.hasNext(); ) {
            RemainingPoints.remove((Integer)i.next());
        }
        ConstructedElements.clear();
    }

    private void Run()  {
        while(ElementsToExecute.size()!= 0) {
            QuickHullStep((HullElement3D)ElementsToExecute.get(0));
        }
    }

    private void addConvexHullNode(point p) {
        if(!ConvexHullNodes.contains(p))
            ConvexHullNodes.add(p);
    }
    
    private void addConvexHullEdge(point p1, point p2) {
        line l;
        int control = 1;
        if(ConvexHullSceleton.size()==0)
            ConvexHullSceleton.add(new line(p1, p2));
        for(int i = 0; i < ConvexHullSceleton.size(); ++i) {
            l = ConvexHullSceleton.get(i);
            if(l.getX1()==p1.getX() && l.getY1()==p1.getY() && l.getZ1()==p1.getZ() && l.getX2()==p2.getX() && l.getY2()==p2.getY() && l.getZ2()==p2.getZ())
                control = 0;
            if(l.getX1()==p2.getX() && l.getY1()==p2.getY() && l.getZ1()==p2.getZ() && l.getX2()==p1.getX() && l.getY2()==p1.getY() && l.getZ2()==p1.getZ())
                control = 0;
        }
        if(control == 1)
            ConvexHullSceleton.add(new line(p1, p2));
    }

    private void addConvexHullPlane(point p1, point p2, point p3) {
        ConvexHullSurface.add(new triangle(p1, p2, p3));
    }

    public ArrayList<point> getConvexHullNodes() {
        Run();
        return ConvexHullNodes;
    }

    public ArrayList<line> getConvexHullSceleton() {
        Run();
        return ConvexHullSceleton;
    }

    public ArrayList<triangle> getConvexHullSurface() {
        Run();
        return ConvexHullSurface;
    }

}
